"""
The xyY color space.

https://en.wikipedia.org/wiki/CIE_1931_color_space#CIE_xy_chromaticity_diagram_and_the_CIE_xyY_color_space
"""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, WHITES
import re
from coloraide import util
from coloraide.util import MutableVector
from typing import Tuple


class XyY(Space):
    """The xyY class."""

    BASE = "xyz-d65"
    NAME = "xyy"
    SERIALIZE = ("--xyy",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("x", "y", "Y")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 1.0)
    )

    @property
    def x(self) -> float:
        """The x channel."""

        return self._coords[0]

    @x.setter
    def x(self, value: float) -> None:
        """Set x."""

        self._coords[0] = value

    @property
    def y(self) -> float:
        """The y channel."""

        return self._coords[1]

    @y.setter
    def y(self, value: float) -> None:
        """Set y."""

        self._coords[1] = value

    @property
    def Y(self) -> float:
        """Y channel."""

        return self._coords[2]

    @Y.setter
    def Y(self, value: float) -> None:
        """Set Y."""

        self._coords[2] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ."""

        return util.xy_to_xyz(coords[0:2], coords[2])

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ."""

        return util.xyz_to_xyY(coords, cls.white())



"""
Prismatic color space.

Creates a Maxwell color triangle with a lightness component.

http://psgraphics.blogspot.com/2015/10/prismatic-color-model.html
https://studylib.net/doc/14656976/the-prismatic-color-space-for-rgb-computations
"""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, GamutBound, WHITES
import re
from coloraide.util import MutableVector
from typing import Tuple


def srgb_to_lrgb(rgb: MutableVector) -> MutableVector:
    """Convert sRGB to Prismatic."""

    l = max(rgb)
    s = sum(rgb)
    return [l] + ([(c / s) for c in rgb] if s != 0 else [0, 0, 0])


def lrgb_to_srgb(lrgb: MutableVector) -> MutableVector:
    """Convert Prismatic to sRGB."""

    rgb = lrgb[1:]
    l = lrgb[0]
    mx = max(rgb)
    return [(l * c) / mx for c in rgb] if mx != 0 else [0, 0, 0]


class Prismatic(Space):
    """The Prismatic color class."""

    BASE = "srgb"
    NAME = "prismatic"
    SERIALIZE = ("--prismatic",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("l", "r", "g", "b")
    CHANNEL_ALIASES = {
        "lightness": 'l',
        "red": 'r',
        "green": 'g',
        "blue": 'b'
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=4))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0)
    )

    @property
    def l(self) -> float:
        """The lightness channel."""

        return self._coords[0]

    @l.setter
    def l(self, value: float) -> None:
        """Set lightness."""

        self._coords[0] = value

    @property
    def r(self) -> float:
        """The red channel."""

        return self._coords[1]

    @r.setter
    def r(self, value: float) -> None:
        """Set red."""

        self._coords[1] = value

    @property
    def g(self) -> float:
        """The green channel."""

        return self._coords[2]

    @g.setter
    def g(self, value: float) -> None:
        """Set green."""

        self._coords[2] = value

    @property
    def b(self) -> float:
        """Y channel."""

        return self._coords[3]

    @b.setter
    def b(self, value: float) -> None:
        """Set Y."""

        self._coords[3] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To sRGB."""

        return lrgb_to_srgb(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From sRGB."""

        return srgb_to_lrgb(coords)



"""
Calculate IPT matricies.

https://scholarworks.rit.edu/cgi/viewcontent.cgi?article=3862&context=theses
"""
import numpy as np

np.set_printoptions(precision=None, sign='-', floatmode='unique')


def xy_to_xyz(x, y):
    """Convert `xyY` to `xyz`."""

    return [x / y, 1, (1 - x - y) / y]


white_d65 = np.asfarray(xy_to_xyz(0.31270, 0.32900))


def white_space_fixup(m):
    """
    Fix up the IPT white space.

    The IPT paper noted that the D65 white point that they use was [0.9504, 1.0, 1.0889].
    This is not the white point we use, so actual precision of the color space breaks down
    and doesn't match implementations that do use the above white point.

    If we look at Colour Science, we can see that when using the above white point, that
    precsion around white is much better:

    ```
    >>> import colour
    >>> import numpy as np
    >>> XYZ = np.array([0.9504, 1.0, 1.0889])
    >>> colour.XYZ_to_IPT(XYZ).tolist()
    [0.9999910919149724, 6.691134620062655e-05, -3.9005477081299755e-05]
    ```

    But what we got was:

    ```
    >>> from coloraide import Color
    >>> from coloraide_extras.spaces.ipt import IPT
    >>> class ColorX(Color): ...
    ...
    >>> ColorX.register(IPT)
    >>> ColorX('white').convert('ipt').coords()
    [1.0000046779854483, 0.00011652905964981697, -0.00010857262923669175]
    ```

    Clearly, this is because of a different white point. By simply removing the
    previous white point and adding ours in, we get comparable results.

    ```
    >>> from coloraide import Color
    >>> from coloraide_extras.spaces.ipt import IPT
    >>> class ColorX(Color): ...
    ...
    >>> ColorX.register(IPT)
    >>> ColorX('white').convert('ipt').coords()
    [0.9999910919149725, 6.69113462008486e-05, -3.900547708157731e-05]
    ```
    """

    theirs = np.diag([0.9504, 1.0, 1.0889])
    ours = np.diag(white_d65)

    return np.dot(np.dot(m, theirs), np.linalg.inv(ours))


m1 = np.asfarray(
    [
        [0.4002, 0.7075, -0.0807],
        [-0.2280, 1.1500, 0.0612],
        [0.0, 0.0, 0.9184]
    ]
)

m1 = white_space_fixup(m1)

m2 = np.asfarray(
    [
        [0.4000, 0.4000, 0.2000],
        [4.4550, -4.8510, 0.3960],
        [0.8056, 0.3572, -1.1628]
    ]
)


if __name__ == "__main__":
    print('===== XYZ to LMS =====')
    print(m1)
    print('===== LMS to XYZ =====')
    print(np.linalg.inv(m1))
    print('===== LMS to IPT =====')
    print(m2)
    print('===== IPT to LMS =====')
    print(np.linalg.inv(m2))



"""
The IPT color space.

https://www.researchgate.net/publication/\
221677980_Development_and_Testing_of_a_Color_Space_IPT_with_Improved_Hue_Uniformity.
"""
from coloraide.spaces import Space, Labish, RE_DEFAULT_MATCH, GamutUnbound, WHITES
import re
from coloraide import util
from coloraide.util import MutableVector
from typing import Tuple, cast

XYZ_TO_LMS = [
    [0.4001764512951712, 0.7075, -0.08068831054981859],
    [-0.2279865839462744, 1.15, 0.061191135138152386],
    [0.0, 0.0, 0.9182669691320122]
]

LMS_TO_XYZ = [
    [1.8503518239760195, -1.1383686221417688, 0.23844898940542367],
    [0.36683077517134854, 0.6438845448402356, -0.010673443584379994],
    [0.0, 0.0, 1.089007917757562]
]

LMS_TO_IPT = [
    [0.4000, 0.4000, 0.2000],
    [4.4550, -4.851, 0.3960],
    [0.8056, 0.3572, -1.1628]
]

IPT_TO_LMS = [
    [0.9999999999999999, 0.09756893051461393, 0.20522643316459166],
    [0.9999999999999999, -0.11387648547314712, 0.1332171583699981],
    [1.0, 0.0326151099170664, -0.6768871830691793]
]


def xyz_to_ipt(xyz: MutableVector) -> MutableVector:
    """XYZ to IPT."""

    lms_p = [util.npow(c, 0.43) for c in cast(MutableVector, util.dot(XYZ_TO_LMS, xyz))]
    return cast(MutableVector, util.dot(LMS_TO_IPT, lms_p))


def ipt_to_xyz(ipt: MutableVector) -> MutableVector:
    """IPT to XYZ."""

    lms = [util.nth_root(c, 0.43) for c in cast(MutableVector, util.dot(IPT_TO_LMS, ipt))]
    return cast(MutableVector, util.dot(LMS_TO_XYZ, lms))


class IPT(Labish, Space):
    """The IPT class."""

    BASE = "xyz-d65"
    NAME = "ipt"
    SERIALIZE = ("--ipt",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("i", "p", "t")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutUnbound(0.0, 1.0),
        GamutUnbound(-0.5, 0.5),
        GamutUnbound(-0.5, 0.5)
    )

    @property
    def i(self) -> float:
        """The `I` channel."""

        return self._coords[0]

    @i.setter
    def i(self, value: float) -> None:
        """Set `I`."""

        self._coords[0] = value

    @property
    def p(self) -> float:
        """The `P` channel."""

        return self._coords[2]

    @p.setter
    def p(self, value: float) -> None:
        """Set `P`."""

        self._coords[2] = value

    @property
    def t(self) -> float:
        """The `T` channel."""

        return self._coords[1]

    @t.setter
    def t(self, value: float) -> None:
        """Set `T`."""

        self._coords[1] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ."""

        return ipt_to_xyz(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ."""

        return xyz_to_ipt(coords)



"""
Uncalibrated, naive CMYK color space.

https://www.w3.org/TR/css-color-5/#cmyk-rgb
"""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, GamutBound, WHITES
import re
from coloraide.util import MutableVector
from typing import Tuple


def srgb_to_cmyk(rgb: MutableVector) -> MutableVector:
    """Convert sRGB to CMYK."""

    k = 1.0 - max(rgb)
    c = m = y = 0.0
    if k != 1:
        r, g, b = rgb
        c = (1.0 - r - k) / (1.0 - k)
        m = (1.0 - g - k) / (1.0 - k)
        y = (1.0 - b - k) / (1.0 - k)

    return [c, m, y, k]


def cmyk_to_srgb(cmyk: MutableVector) -> MutableVector:
    """Convert CMYK to sRGB."""

    c, m, y, k = cmyk
    return [
        1.0 - min(1.0, c * (1.0 - k) + k),
        1.0 - min(1.0, m * (1.0 - k) + k),
        1.0 - min(1.0, y * (1.0 - k) + k)
    ]


class CMYK(Space):
    """The CMYK color class."""

    BASE = "srgb"
    NAME = "cmyk"
    SERIALIZE = ("--cmyk",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("c", "m", "y", "k")
    CHANNEL_ALIASES = {
        "cyan": 'c',
        "magenta": 'm',
        "yellow": 'y',
        "black": 'k'
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=4))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0)
    )

    @property
    def c(self) -> float:
        """The cyan channel."""

        return self._coords[0]

    @c.setter
    def c(self, value: float) -> None:
        """Set cyan."""

        self._coords[0] = value

    @property
    def m(self) -> float:
        """The magenta channel."""

        return self._coords[1]

    @m.setter
    def m(self, value: float) -> None:
        """Set magenta."""

        self._coords[1] = value

    @property
    def y(self) -> float:
        """The yellow channel."""

        return self._coords[2]

    @y.setter
    def y(self, value: float) -> None:
        """Set yellow."""

        self._coords[2] = value

    @property
    def k(self) -> float:
        """The black channel."""

        return self._coords[3]

    @k.setter
    def k(self, value: float) -> None:
        """Set black."""

        self._coords[3] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To sRGB."""

        return cmyk_to_srgb(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From sRGB."""

        return srgb_to_cmyk(coords)



"""ColorAide Extra."""
from .__meta__ import __version_info__, __version__  # noqa: F401
from coloraide import Color as Base
from .spaces.hsi import HSI
from .spaces.ipt import IPT
from .spaces.igpgtg import IgPgTg
from .spaces.cmy import CMY
from .spaces.cmyk import CMYK
from .spaces.ucs import UCS
from .spaces.uvw import UVW
from .spaces.xyy import XyY
from .spaces.hunter_lab import HunterLab
from .spaces.prismatic import Prismatic

__all__ = ("Color", "SPACES")

SPACES = [HSI, IPT, IgPgTg, CMY, CMYK, UCS, UVW, XyY, HunterLab, Prismatic]


class Color(Base):
    """Color class containing all default and extra color spaces."""


Color.register(SPACES)



"""HSI class."""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, FLG_ANGLE, FLG_OPT_PERCENT, GamutBound, Cylindrical, WHITES
from coloraide import util
import re
from coloraide.util import MutableVector
from typing import Tuple


def srgb_to_hsi(rgb: MutableVector) -> MutableVector:
    """SRGB to HSI."""

    r, g, b = rgb
    h = util.NaN
    s = 0.0
    mx = max(rgb)
    mn = min(rgb)
    i = sum(rgb) * 1 / 3
    s = 0 if i == 0 else 1 - (mn / i)
    c = mx - mn

    if c != 0.0:
        if mx == r:
            h = (g - b) / c
        elif mx == g:
            h = (b - r) / c + 2.0
        else:
            h = (r - g) / c + 4.0
        h *= 60.0

    return [util.constrain_hue(h), s, i]


def hsi_to_srgb(hsi: MutableVector) -> MutableVector:
    """
    HSI to RGB.

    https://en.wikipedia.org/wiki/HSL_and_HSV#Saturation
    """

    h, s, i = hsi
    h = h % 360
    h /= 60
    z = 1 - abs(h % 2 - 1)
    c = (3 * i * s) / (1 + z)
    x = c * z

    if util.is_nan(h):
        rgb = [0.0] * 3
    elif 0 <= h <= 1:
        rgb = [c, x, 0]
    elif 1 <= h <= 2:
        rgb = [x, c, 0]
    elif 2 <= h <= 3:
        rgb = [0, c, x]
    elif 3 <= h <= 4:
        rgb = [0, x, c]
    elif 4 <= h <= 5:
        rgb = [x, 0, c]
    else:
        rgb = [c, 0, x]
    m = i * (1 - s)

    return [chan + m for chan in rgb]


class HSI(Cylindrical, Space):
    """HSI class."""

    BASE = "srgb"
    NAME = "hsi"
    SERIALIZE = ("--hsi",)
    CHANNEL_NAMES = ("h", "s", "i")
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "intensity": "i"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']
    GAMUT_CHECK = "srgb"

    BOUNDS = (
        GamutBound(0.0, 360.0, FLG_ANGLE),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT)
    )

    @property
    def h(self) -> float:
        """Hue channel."""

        return self._coords[0]

    @h.setter
    def h(self, value: float) -> None:
        """Shift the hue."""

        self._coords[0] = value

    @property
    def s(self) -> float:
        """Saturation channel."""

        return self._coords[1]

    @s.setter
    def s(self, value: float) -> None:
        """Set saturation."""

        self._coords[1] = value

    @property
    def i(self) -> float:
        """Intensity channel."""

        return self._coords[2]

    @i.setter
    def i(self, value: float) -> None:
        """Set intensity channel."""

        self._coords[2] = value

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[2] == 0:
            coords[0] = util.NaN

        return coords, alpha

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To sRGB from HSI."""

        return hsi_to_srgb(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From sRGB to HSI."""

        return srgb_to_hsi(coords)



"""Uncalibrated, naive CMY color space."""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, GamutBound, WHITES
import re
from coloraide.util import MutableVector
from typing import Tuple


def srgb_to_cmy(rgb: MutableVector) -> MutableVector:
    """Convert sRGB to CMY."""

    return [1 - c for c in rgb]


def cmy_to_srgb(cmy: MutableVector) -> MutableVector:
    """Convert cmy to sRGB."""

    return [1 - c for c in cmy]


class CMY(Space):
    """The CMY color class."""

    BASE = "srgb"
    NAME = "cmy"
    SERIALIZE = ("--cmy",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("c", "m", "y")
    CHANNEL_ALIASES = {
        "cyan": 'c',
        "magenta": 'm',
        "yellow": 'y'
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=4))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0),
        GamutBound(0.0, 1.0)
    )

    @property
    def c(self) -> float:
        """The cyan channel."""

        return self._coords[0]

    @c.setter
    def c(self, value: float) -> None:
        """Set cyan."""

        self._coords[0] = value

    @property
    def m(self) -> float:
        """The magenta channel."""

        return self._coords[1]

    @m.setter
    def m(self, value: float) -> None:
        """Set magenta."""

        self._coords[1] = value

    @property
    def y(self) -> float:
        """The yellow channel."""

        return self._coords[2]

    @y.setter
    def y(self, value: float) -> None:
        """Set yellow."""

        self._coords[2] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To sRGB."""

        return cmy_to_srgb(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From sRGB."""

        return srgb_to_cmy(coords)



"""
CIE 1960 UCS color class.

http://en.wikipedia.org/wiki/CIE_1960_color_space#Relation_to_CIE_XYZ
"""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, WHITES
import re
from coloraide.util import MutableVector
from typing import Tuple


def xyz_to_ucs(xyz: MutableVector) -> MutableVector:
    """Translate XYZ to 1960 UCS."""

    x, y, z = xyz
    return [(2 / 3) * x, y, (-x + 3 * y + z) * 0.5]


def ucs_to_xyz(ucs: MutableVector) -> MutableVector:
    """Translate 1960 UCS to XYZ."""

    u, v, w = ucs
    return [(3 / 2) * u, v, (3 / 2) * u - 3 * v + 2 * w]


class UCS(Space):
    """The 1960 UCS class."""

    BASE = "xyz-d65"
    NAME = "ucs"
    SERIALIZE = ("--ucs",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("u", "v", "w")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 1.0)
    )

    @property
    def u(self) -> float:
        """The u channel."""

        return self._coords[0]

    @u.setter
    def u(self, value: float) -> None:
        """Set u."""

        self._coords[0] = value

    @property
    def v(self) -> float:
        """The v channel."""

        return self._coords[1]

    @v.setter
    def v(self, value: float) -> None:
        """Set v."""

        self._coords[1] = value

    @property
    def w(self) -> float:
        """The w channel."""

        return self._coords[2]

    @w.setter
    def w(self, value: float) -> None:
        """Set w."""

        self._coords[2] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ."""

        return ucs_to_xyz(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ."""

        return xyz_to_ucs(coords)



"""
CIE 1964 UVW class.

https://en.wikipedia.org/wiki/CIE_1964_color_space
"""
from coloraide.spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, WHITES
import re
from coloraide import util
from coloraide.util import MutableVector, Vector
from typing import Tuple


def xyz_to_uvw(xyz: MutableVector, white: Vector) -> MutableVector:
    """
    Translate XYZ to UVW.

    When translating XYZ to the intermediate xyY, we need to treat Y as Y * 100.
    """

    u, v = util.xy_to_uv_1960(util.xyz_to_xyY(xyz, white)[:2])
    u0, v0 = util.xy_to_uv_1960(white)
    w = 25.0 * util.nth_root(xyz[1] * 100.0, 3) - 17.0
    return [
        13 * w * (u - u0),
        13 * w * (v - v0),
        w
    ]


def uvw_to_xyz(uvw: MutableVector, white: Vector) -> MutableVector:
    """
    Translate UVW to XYZ.

    When translating xyY back to XYZ, we need to scale Y back as well: Y / 100.
    """

    u0, v0 = util.xy_to_uv_1960(white)
    w = uvw[2]
    x, y = util.uv_1960_to_xy(
        [
            (uvw[0] / (13 * w)) + u0,
            (uvw[1] / (13 * w)) + v0
        ]
    )
    return util.xy_to_xyz([x, y], (((w + 17.0) / 25.0) ** 3) / 100.0)


class UVW(Space):
    """The UVW class."""

    BASE = "xyz-d65"
    NAME = "uvw"
    SERIALIZE = ("--uvw",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("u", "v", "w")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutUnbound(-0.1, 1.0),
        GamutUnbound(-0.1, 1.0),
        GamutUnbound(0.0, 1.0)
    )

    @property
    def u(self) -> float:
        """The u channel."""

        return self._coords[0]

    @u.setter
    def u(self, value: float) -> None:
        """Set u."""

        self._coords[0] = value

    @property
    def v(self) -> float:
        """The v channel."""

        return self._coords[1]

    @v.setter
    def v(self, value: float) -> None:
        """Set v."""

        self._coords[1] = value

    @property
    def w(self) -> float:
        """The w channel."""

        return self._coords[2]

    @w.setter
    def w(self, value: float) -> None:
        """Set w."""

        self._coords[2] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ."""

        return uvw_to_xyz(coords, cls.white())

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ."""

        return xyz_to_uvw(coords, cls.white())



"""
The IgPgTg color space.

https://www.ingentaconnect.com/content/ist/jpi/2020/00000003/00000002/art00002#
"""
from coloraide.spaces import Space, Labish, RE_DEFAULT_MATCH, GamutUnbound, WHITES
import re
from coloraide import util
from coloraide.util import MutableVector
from typing import Tuple, cast

XYZ_TO_LMS = [
    [2.968, 2.741, -0.649],
    [1.237, 5.969, -0.173],
    [-0.318, 0.387, 2.311]
]

LMS_TO_XYZ = [
    [0.43434868555746337, -0.20636237011428415, 0.10653033617352774],
    [-0.08785463778363381, 0.20846346647992342, -0.009066845616854866],
    [0.07447971736457795, -0.06330532030466152, 0.44889031421761344]
]

LMS_TO_IGPGTG = [
    [0.117, 1.464, 0.130],
    [8.285, -8.361, 21.40],
    [-1.208, 2.412, -36.53]
]

IGPGTG_TO_LMS = [
    [0.581846461899246, 0.12331854793907822, 0.07431308420320765],
    [0.634548193791416, -0.009437923746683554, -0.003270744675229782],
    [0.02265698651657832, -0.004701151874826367, -0.030048158824914562]
]


def xyz_to_igpgtg(xyz: MutableVector) -> MutableVector:
    """XYZ to IgPgTg."""

    lms_in = cast(MutableVector, util.dot(XYZ_TO_LMS, xyz))
    lms = [
        util.npow(lms_in[0] / 18.36, 0.427),
        util.npow(lms_in[1] / 21.46, 0.427),
        util.npow(lms_in[2] / 19435, 0.427)
    ]
    return cast(MutableVector, util.dot(LMS_TO_IGPGTG, lms))


def igpgtg_to_xyz(itp: MutableVector) -> MutableVector:
    """IgPgTg to XYZ."""

    lms = cast(MutableVector, util.dot(IGPGTG_TO_LMS, itp))
    lms_in = [
        util.nth_root(lms[0], 0.427) * 18.36,
        util.nth_root(lms[1], 0.427) * 21.46,
        util.nth_root(lms[2], 0.427) * 19435
    ]
    return cast(MutableVector, util.dot(LMS_TO_XYZ, lms_in))


class IgPgTg(Labish, Space):
    """The IgPgTg class."""

    BASE = "xyz-d65"
    NAME = "igpgtg"
    SERIALIZE = ("--igpgtg",)  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("ig", "tg", "pg")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']

    BOUNDS = (
        GamutUnbound(0.0, 1.0),
        GamutUnbound(-0.5, 0.5),
        GamutUnbound(-0.5, 0.5)
    )

    @property
    def ig(self) -> float:
        """The `Ig` channel."""

        return self._coords[0]

    @ig.setter
    def ig(self, value: float) -> None:
        """Set `Ig`."""

        self._coords[0] = value

    @property
    def tg(self) -> float:
        """The `Tg` channel."""

        return self._coords[1]

    @tg.setter
    def tg(self, value: float) -> None:
        """Set `Tg`."""

        self._coords[1] = value

    @property
    def pg(self) -> float:
        """The `Pg` channel."""

        return self._coords[2]

    @pg.setter
    def pg(self, value: float) -> None:
        """Set `Pg`."""

        self._coords[2] = value

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ."""

        return igpgtg_to_xyz(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ."""

        return xyz_to_igpgtg(coords)



"""
Hunter Lab class.

https://support.hunterlab.com/hc/en-us/articles/203997095-Hunter-Lab-Color-Scale-an08-96a2
"""
from coloraide.spaces import RE_DEFAULT_MATCH, WHITES
from coloraide.spaces.lab import Lab
from coloraide import util
import re
from coloraide.util import MutableVector
from typing import cast

# Values for the original Hunter Lab with illuminant C.
# Also, the original calculated `ka` and `kb` under illuminant C.
# Original values:
#   https://support.hunterlab.com/hc/en-us/articles/203997095-Hunter-Lab-Color-Scale-an08-96a2
# Slightly more precise values in `rdab`:
#   https://hunterlabdotcom.files.wordpress.com/2012/07/an-1016-hunter-rd-a-b-color-scale-update-12-07-03.pdf
CXN = 98.043
CYN = 100.0
CZN = 118.106
CKA = 175.0
CKB = 70.0
# Conversion factors
CKA_FACTOR = CKA / (CXN + CYN)
CKB_FACTOR = CKB / (CZN + CYN)


def xyz_to_hlab(xyz: MutableVector, white: MutableVector) -> MutableVector:
    """Convert XYZ to Hunter Lab."""

    xn, yn, zn = cast(MutableVector, util.multiply(util.xy_to_xyz(white), 100))
    ka = CKA_FACTOR * (xn + yn)
    kb = CKB_FACTOR * (yn + zn)
    x, y, z = cast(MutableVector, util.multiply(xyz, 100))
    l = util.nth_root(y / yn, 2)
    a = b = 0
    if l != 0:
        a = ka * (x / xn - y / yn) / l
        b = kb * (y / yn - z / zn) / l
    return [l * 100, a, b]


def hlab_to_xyz(hlab: MutableVector, white: MutableVector) -> MutableVector:
    """Convert Hunter Lab to XYZ."""

    xn, yn, zn = cast(MutableVector, util.multiply(util.xy_to_xyz(white), 100))
    ka = CKA_FACTOR * (xn + yn)
    kb = CKB_FACTOR * (yn + zn)
    l, a, b = hlab
    l /= 100
    y = (l ** 2) * yn
    x = (((a * l) / ka) + (y / yn)) * xn
    z = (((b * l) / kb) - (y / yn)) * -zn
    return cast(MutableVector, util.divide([x, y, z], 100))


class HunterLab(Lab):
    """Hunter Lab class."""

    BASE = 'xyz-d65'
    NAME = "hunter-lab"
    SERIALIZE = ("--hunter-lab",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = WHITES['2deg']['D65']

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from Hunter Lab."""

        return hlab_to_xyz(coords, cls.white())

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to Hunter Lab."""

        return xyz_to_hlab(coords, cls.white())
